from functools import wraps


_hooks_before = {}
_hooks_after = {}


def Hook(function=None, id=None):
    """Hook decorator
    Use to decorate functions as hooks, so plugins can hook up their custom functions.
    """
    # `id` passed as `arg` not `kwarg`
    if isinstance(function, str):
        return Hook(id=function)

    def decorator(function):
        @wraps(function)
        def wrapped(*args, **kwargs):
            _id = id if id is not None else function.__qualname__
            # Hooks before
            for f in _hooks_before.get(_id, []):
                f(*args, **kwargs)
            # Main function
            result = function(*args, **kwargs)
            # Hooks after
            for f in _hooks_after.get(_id, []):
                f(*args, hook_result=result, **kwargs)
            return result

        return wrapped

    # Called @Hook or we are in the second step
    if callable(function):
        return decorator(function)
    else:
        return decorator


def HookBefore(id: str):
    """Decorator for functions to be called before a Hook-Function is called
    The hooked up function must accept the same arguments as the function hooked onto,
    as the functions are called with the same arguments.
    Hint: This enables you to modify the arguments!
    """
    if not id or not isinstance(id, str):
        raise TypeError("HookBefore requires the ID of the function to hook up")

    def wrapped(function):
        _hooks_before.setdefault(id, []).append(function)
        return function

    return wrapped


def HookAfter(id: str):
    """Decorator for functions to be called after a Hook-Function is called
    As with the HookBefore, the hooked up function must accept the same
    arguments as the function hooked onto, but also receives a
    `hook_result` kwarg containing the result of the function.
    """

    if not id or not isinstance(id, str):
        raise TypeError("HookAfter requires the ID of the function to hook up")

    def wrapped(function):
        _hooks_after.setdefault(id, []).append(function)
        return function

    return wrapped
